/*
 
 Copyright (C) 2006            Jacek Piszczek (fuhQuakeLauncher)
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 
 See the included (GNU.txt) GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <unistd.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#import "myController.h"

static int sysctlbyname_with_pid (const char *name, pid_t pid, 
                            void *oldp, size_t *oldlenp, 
                            void *newp, size_t newlen)
{
    if (pid == 0) 
    {
        if (sysctlbyname(name, oldp, oldlenp, newp, newlen) == -1)  
        {
            return -1;
        }
    } 
    else 
    {
        int mib[CTL_MAXNAME];
        size_t len = CTL_MAXNAME;
        if (sysctlnametomib(name, mib, &len) == -1) 
        {
            return -1;
        }
        mib[len] = pid;
        len++;
        if (sysctl(mib, len, oldp, oldlenp, newp, newlen) == -1)  
        {
            return -1;
        }
    }
    return 0;
}

// checks if a process is run using Rosetta (in order to run the wintel bin)
// returns 1 if we're run on a real PowerPC Mac
int is_pid_native(pid_t pid)
{
    int ret = 0;
    size_t sz = sizeof(ret);
 
    if (sysctlbyname_with_pid("sysctl.proc_native", pid, 
                &ret, &sz, NULL, 0) == -1)
    {
        if (errno == ENOENT)
        {
            return 1;
        }
        return -1;
    }
    return ret;
}

@implementation myController

- (IBAction)selectbasedir:(id)sender
{
    chdir([[[NSBundle mainBundle] bundlePath] UTF8String]);

    NSOpenPanel *panel = [NSOpenPanel openPanel];
    NSString *cpath = [[basedir stringValue] stringByStandardizingPath];
    
    if ((cpath == nil) || [cpath length] < 1) 
    {
        cpath = [NSString stringWithUTF8String:"../basedir"];
    }
    
    if (![cpath isAbsolutePath])
    {
        // we need an abs path here, damnit
        char buffer[4096];
        
        getcwd(buffer,4096);
        
        NSString *cwd = [NSString stringWithUTF8String:buffer];
        cwd = [cwd stringByAppendingPathComponent:cpath];
        
        cpath = [cwd stringByStandardizingPath];
    }
    
    NSString *parent = [cpath stringByDeletingLastPathComponent];
    NSString *directory = [cpath lastPathComponent];

    [panel setCanChooseFiles:NO];
    [panel setCanChooseDirectories:YES];

    if ([panel runModalForDirectory:parent file:directory types:nil])
    {
        NSArray *dirarray = [panel filenames];
        NSString *dir = [dirarray objectAtIndex:0];
        
        if (dir != nil)
        {
            // wtf don't the fucking cocoa bindings work properly here??
            // can somebody fucking tell me wtf isn't this change saved??
            // argh!
            [basedir setStringValue:dir];
        }
    }
}

#define MAX_ARGS 64

- (IBAction)start:(id)sender
{
    // NOTE: config name is not converted, it cannot contain non-ascii chars!
    NSFileManager *dman = [NSFileManager defaultManager];
    char basedir_path[4096] = "../basedir";
    char config_file[128] = "config";
    char params_string[512] = "";
    char *paramsstr = params_string;
    void *argv[64 + 3];
    int lastarg = 0;
    BOOL defaultbdir = true;
    BOOL alternativeconfig = false;

    char *bdir_user = (char*)[[basedir stringValue] UTF8String];
    char *config_user = (char*)[[config stringValue] UTF8String];
    char *width_user = (char*)[[width stringValue] UTF8String];
    char *height_user = (char*)[[height stringValue] UTF8String];
    char *depth_user = 0;
    char *params_user = (char*)[[params stringValue] UTF8String];
    
    switch ([depth indexOfSelectedItem])
    {
        case 0: depth_user = "15"; break;
        case 1: depth_user = "16"; break;
        case 2: default: depth_user = "24"; break;
    }

    // check if bdir_user was set to anything at all
    if ((bdir_user != nil) && bdir_user[0])
    {
        strlcpy(basedir_path,
                [dman fileSystemRepresentationWithPath:[basedir stringValue]],
                4096);
        defaultbdir = false;
    }

    if ((config_user != nil) && config_user[0])
    {
        strlcpy(config_file,config_user,128);
        alternativeconfig = true;
    }

    if ((params_user != nil) && params_user[0])
    {
        // leave 1 byte free, plz
        strlcpy(params_string,params_user,511);
    }

    // build std args

    // is fuhQuakeLoader running on a PowerPC machine?
    if (is_pid_native(getpid()))
    {
        argv[lastarg++] = "./Contents/Resources/fuhQuake.gl";
    }
    else
    {
        // run the wintel binary if the loader is translated by Rosetta
        argv[lastarg++] = "./Contents/Resources/fuhQuake.gl.i386";
    }
    
    argv[lastarg++] = "-basedir";
    argv[lastarg++] = basedir_path;
    
    if (alternativeconfig)
    {
        argv[lastarg++] = "+cfg_load";
        argv[lastarg++] = config_file;
    }

    if ((width_user != 0) && (height_user != 0) && (width_user[0]))
    {
        argv[lastarg++] = "-width";
        argv[lastarg++] = width_user;
        
        argv[lastarg++] = "-height";
        argv[lastarg++] = height_user;
    }
    
    if ((depth_user != 0) && (depth_user[0]))
    {
        argv[lastarg++] = "-depth";
        argv[lastarg++] = depth_user;
    }

    // look for additional arguments
    int plength = 0;
    
    // lame params string analyzer :)
    // doesn't handle "" cause I'm too lazy. add it if you need it :)
    do
    {
        paramsstr++;
        plength++;
        
        if ((*paramsstr == ' ') || (*paramsstr == 0))
        {
            *paramsstr = 0;
            argv[lastarg++] = paramsstr - plength;
            plength = 0;
            paramsstr++;
        }
        
        if (lastarg > MAX_ARGS) break;
    } while (*paramsstr);

    // terminate args
    argv[lastarg] = nil;

    // make sure we're in a proper current dir
    chdir([dman fileSystemRepresentationWithPath:
          [[NSBundle mainBundle] bundlePath]]);

    // check it the pak0.pak/pak1.pak files are in the correct location
    // note: make sure to support default setting (basedir contains "")
    NSString *pakopath1 = (defaultbdir ? @"../basedir/id1/pak0.pak" : 
        [[basedir stringValue] stringByAppendingPathComponent:@"id1/pak0.pak"]);
    NSString *pakopath2 = (defaultbdir ? @"../basedir/id1/pak1.pak" : 
        [[basedir stringValue] stringByAppendingPathComponent:@"id1/pak1.pak"]);

    int pakf1 = open([dman fileSystemRepresentationWithPath:pakopath1],
                     O_RDONLY,0);
    int pakf2 = open([dman fileSystemRepresentationWithPath:pakopath2],
                     O_RDONLY,0);

    if ((pakf1 == -1) || (pakf2 == -1))
    {
        [[NSAlert 
            alertWithMessageText:
                @"fuhQuake is unable to locate the pak files!"
            defaultButton: @"OK"
            alternateButton: @"" otherButton:@""
            informativeTextWithFormat:
                @"Please check if pak0.pak and pak1.pak are available in "
                "the id1 folder of basedir and that the basedir path "
                "is correctly configured in Preferences"
            ] runModal];
        if (pakf1 != -1) close(pakf1);
        if (pakf2 != -1) close(pakf2);
        return; // do not launch fuhQ!
    }

    if (pakf1 != -1) close(pakf1);
    if (pakf2 != -1) close(pakf2);

    // fork and launch
    if (fork() == 0)
    {
        execvp(argv[0],(char**)argv);

        exit(0);
    }
}

// sent when the user wants to close the main window (terminate the app)
- (void)windowWillClose:(NSNotification *)aNotification
{
    [NSApp terminate:self];
}

@end
